==> audiochain-1.1.5-src.tgz: org/audiochain/devices/drum/DrumAudioDevice.java <==
/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.drum;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;

import javax.sound.sampled.UnsupportedAudioFileException;
import javax.xml.parsers.ParserConfigurationException;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.FrameObserver;
import org.audiochain.io.LimitingAudioDataReader;
import org.audiochain.model.AbstractAudioDevice;
import org.audiochain.model.AudioDeviceProperty;
import org.audiochain.model.BinaryAudioProject;
import org.audiochain.model.BinaryDataset;
import org.audiochain.model.BinaryFileDataset;
import org.audiochain.model.BinarySynchronizable;
import org.audiochain.model.Performable;
import org.audiochain.model.SourceAudioDevice;
import org.audiochain.model.Synchronizable;
import org.audiochain.model.UserInterfaceContext;
import org.xml.sax.SAXException;

/**
 * TODO remember all synced but unsaved files and remove them when the project
 * becomes closed without saving (analogous to recording device)
 */
public class DrumAudioDevice extends AbstractAudioDevice implements SourceAudioDevice, BinarySynchronizable {
	private static final long serialVersionUID = 1L;
	private static final String DRUM_SCHEMA_PROPERTY_NAME = "DrumSchema";
	private static final String DRUM_RHYTHM_PROPERTY_NAME = "Rhythm";
	private String drumSchemaFileName;
	private String drumRhythmName;
	@AudioDeviceProperty(name = "BPM", min = 20, max = 350, symbol = 9684, unit = "Bpm")
	private float beatsPerMinute = 80F;
	@AudioDeviceProperty(name = "Repeat", symbol = 10226, falseSymbol = 10230)
	private boolean repeat;
	private transient DrumAudioDataReader drumReader;
	private transient Collection<DrumAudioDeviceListener> listeners;
	private transient DrumSchema cachedDrumSchema;

	@Override
	public AudioDataReader createAudioDataReader(float frameRate, int sampleSizeInBits, FrameObserver frameObserver) {
		DrumSchema drumSchema = getDrumSchema();
		DrumRhythm rhythm = drumSchema.getDrumRhythmByName(drumRhythmName);
		drumReader = new DrumAudioDataReader(drumSchema, rhythm, beatsPerMinute, frameRate);
		drumReader.setRepeat(repeat);
		AudioDataReader reader = new LimitingAudioDataReader(drumReader, sampleSizeInBits);
		return reader;
	}

	/*
	 * This method is called from ui.
	 */
	void setDrumSchemaFile(File drumSchemaFile) throws FileNotFoundException, IOException, DrumSchemaFormatException, UnsupportedAudioFileException, SAXException, ParserConfigurationException {
		if (drumSchemaFile == null) {
			throw new IllegalArgumentException("Drum schema file must not be null.");
		}
		// throw exceptions if file is not valid
		DrumSchema schema = loadDrumSchema(drumSchemaFile);
		// update the new file change time and the ui
		String oldDrumSchemaFileName = this.drumSchemaFileName;
		String newDrumSchemaFileName = audioProject.getRelativePath(drumSchemaFile.getAbsolutePath());
		if (!isEqual(oldDrumSchemaFileName, newDrumSchemaFileName)) {
			this.drumSchemaFileName = newDrumSchemaFileName;
			setAudioDeviceValueLastChangeTime(DRUM_SCHEMA_PROPERTY_NAME, System.currentTimeMillis());
			// preselect the first rhythm for the new file
			Collection<String> drumRhythmNames = schema.getDrumRhythmNames();
			// a schema must have at least one rhytm
			this.drumRhythmName = drumRhythmNames.iterator().next();
			setAudioDeviceValueLastChangeTime(DRUM_RHYTHM_PROPERTY_NAME, System.currentTimeMillis());
			cachedDrumSchema = null;
			fireDrumSchemaFileNameChanged();
			fireDrumRhythmChanged();
			firePerformanceRelevantAudioDevicePropertyChanged();
			audioProject.changeFrameLength();
			// trigger the binary drum file synchronization
			fireBinaryChanged(new BinaryAudioProject(audioProject));
		}
	}

	private void fireDrumSchemaFileNameChanged() {
		if (listeners == null) {
			return;
		}
		for (DrumAudioDeviceListener listener : listeners) {
			listener.drumSchemaFileNameChanged(drumSchemaFileName);
		}
	}

	/*
	 * This method must be called only if isAudioDataAvailable() returns true.
	 */
	private DrumSchema getDrumSchema() {
		if (!isAudioDataAvailable()) {
			throw new IllegalStateException("Drum schema is not available.");
		}
		return cachedDrumSchema;
	}

	private DrumSchema loadDrumSchema(File drumSchemaFile) throws FileNotFoundException, IOException, DrumSchemaFormatException, UnsupportedAudioFileException, SAXException,
			ParserConfigurationException {
		DrumSchema ret;
		String projectPath = audioProject.getProjectPath();
		if (drumSchemaFile.getName().endsWith(".h2song")) {
			FileInputStream fis = new FileInputStream(drumSchemaFile);
			BufferedInputStream bis = new BufferedInputStream(fis);
			Collection<File> clipSearchPaths = DrumSettings.getInstance().getClipSearchPaths();
			clipSearchPaths.add(new File(projectPath));
			try {
				HydrogenDrumSchemaReader hydrogenDrumSchemaReader = new HydrogenDrumSchemaReader(bis, clipSearchPaths);
				ret = hydrogenDrumSchemaReader.getSchema();
			} finally {
				bis.close();
				fis.close();
			}
		} else {
			FileReader fr = new FileReader(drumSchemaFile);
			// search for clip files within the project and relative to the drum
			// schema file
			ret = DrumSchema.readDrumSchema(fr, new File(projectPath), drumSchemaFile.getParentFile());
			fr.close();
		}
		return ret;
	}

	/* is called from ui */
	void setDrumRhythm(String drumRhythmName) {
		String oldDrumRhythmName = this.drumRhythmName;
		if (!isEqual(oldDrumRhythmName, drumRhythmName)) {
			this.drumRhythmName = drumRhythmName;
			setAudioDeviceValueLastChangeTime(DRUM_RHYTHM_PROPERTY_NAME, System.currentTimeMillis());
			updateReader();
			firePerformanceRelevantAudioDevicePropertyChanged();
			audioProject.changeFrameLength();
		}
	}

	private void updateReader() {
		if (drumReader != null && isAudioDataAvailable()) {
			DrumSchema drumSchema = getDrumSchema();
			DrumRhythm rhythm = drumSchema.getDrumRhythmByName(drumRhythmName);
			drumReader.setDrumSchemaAndRhythm(drumSchema, rhythm);
		}
	}

	private void fireDrumRhythmChanged() {
		if (listeners == null) {
			return;
		}
		for (DrumAudioDeviceListener listener : listeners) {
			listener.drumRhythmNameChanged(drumRhythmName);
		}
	}

	void setBeatsPerMinute(float beatPerMinute) {
		this.beatsPerMinute = beatPerMinute;
		if (drumReader != null) {
			drumReader.setPulsesPerMinute(beatPerMinute);
		}
		firePerformanceRelevantAudioDevicePropertyChanged();
		audioProject.changeFrameLength();
	}

	void setRepeat(boolean repeat) {
		this.repeat = repeat;
		if (drumReader != null) {
			drumReader.setRepeat(repeat);
		}
		firePerformanceRelevantAudioDevicePropertyChanged();
		audioProject.changeFrameLength();
	}

	Collection<String> getDrumRhythmNames() {
		Collection<String> ret = new LinkedHashSet<String>();
		if (isAudioDataAvailable()) {
			DrumSchema drumSchema = getDrumSchema();
			ret.addAll(drumSchema.getDrumRhythmNames());
		}
		return ret;
	}

	@Override
	public long getMaxFrame() {
		if (!isAudioDataAvailable()) {
			return -1;
		}
		DrumSchema drumSchema = getDrumSchema();
		DrumRhythm rhythm = drumSchema.getDrumRhythmByName(drumRhythmName);
		DrumAudioDataReader drumReader = new DrumAudioDataReader(drumSchema, rhythm, beatsPerMinute, audioProject.getFrameRate());
		drumReader.setRepeat(repeat);
		return drumReader.getRhythmFrameCount();
	}

	@Override
	protected UserInterfaceContext<DrumAudioDevice> createUserInterfaceContext() {
		return new DrumUserInterfaceContext();
	}

	@Override
	public boolean isRecordable() {
		return false;
	}

	@Override
	public Performable getPerformable() {
		return null;
	}

	@Override
	public String getName() {
		return "Drum";
	}

	@Override
	public Collection<BinaryDataset> getAllBinaryDatasets() {
		Collection<BinaryDataset> ret = new HashSet<BinaryDataset>();
		if (drumSchemaFileName != null) {
			File schemaFile = audioProject.getRelativeFile(drumSchemaFileName);
			if (schemaFile.exists()) {
				ret.add(new BinaryFileDataset(audioProject, schemaFile));
			}
		}
		if (isAudioDataAvailable()) {
			DrumSchema drumSchema = getDrumSchema();
			for (DrumClip clip : drumSchema.getClips()) {
				File clipFile = clip.getFile();
				if (clipFile.exists() && clipFile.canRead()) {
					ret.add(new BinaryFileDataset(audioProject, clipFile));
				}
			}
		}
		return ret;
	}

	@Override
	public void synchronizeBy(Synchronizable source) {
		// sync bpm and repeat
		super.synchronizeBy(source);

		if (source instanceof DrumAudioDevice) {
			DrumAudioDevice sourceDevice = (DrumAudioDevice) source;
			long sourceRhythmChangeTime = sourceDevice.getAudioDeviceValueLastChangeTime(DRUM_RHYTHM_PROPERTY_NAME);
			long rhythmChangeTime = getAudioDeviceValueLastChangeTime(DRUM_RHYTHM_PROPERTY_NAME);
			if (!isEqual(sourceDevice.drumSchemaFileName, drumSchemaFileName)) {
				long sourceSchemaChangeTime = sourceDevice.getAudioDeviceValueLastChangeTime(DRUM_SCHEMA_PROPERTY_NAME);
				long schemaChangeTime = getAudioDeviceValueLastChangeTime(DRUM_SCHEMA_PROPERTY_NAME);
				if (greaterOrBothZero(sourceSchemaChangeTime, schemaChangeTime)) {
					DrumSchema schema = loadDrumSchemaIfAvailable(sourceDevice.drumSchemaFileName, sourceDevice.drumRhythmName);
					if (schema != null) {
						drumSchemaFileName = sourceDevice.drumSchemaFileName;
						setAudioDeviceValueLastChangeTime(DRUM_SCHEMA_PROPERTY_NAME, sourceSchemaChangeTime);
						drumRhythmName = sourceDevice.drumRhythmName;
						setAudioDeviceValueLastChangeTime(DRUM_RHYTHM_PROPERTY_NAME, sourceRhythmChangeTime);
						cachedDrumSchema = schema;
						updateReader();
						fireDrumSchemaFileNameChanged();
						fireDrumRhythmChanged();
						firePerformanceRelevantAudioDevicePropertyChanged();
						audioProject.changeFrameLength();
					} else {
						// the schema changed, but is not known at this side
						drumSchemaFileName = null;
						cachedDrumSchema = null;
						// tell the other side that this side needs the binary files
						fireBinaryChanged(new BinaryAudioProject(audioProject));
					}
				}
			} else if (!isEqual(sourceDevice.drumRhythmName, drumRhythmName)) {
				if (greaterOrBothZero(sourceRhythmChangeTime, rhythmChangeTime)) {
					this.drumRhythmName = sourceDevice.drumRhythmName;
					setAudioDeviceValueLastChangeTime(DRUM_RHYTHM_PROPERTY_NAME, sourceRhythmChangeTime);
					updateReader();
					fireDrumRhythmChanged();
					firePerformanceRelevantAudioDevicePropertyChanged();
					audioProject.changeFrameLength();
				}
			}
		}
	}

	@Override
	public boolean isAudioDataAvailable() {
		if (cachedDrumSchema != null) {
			return true;
		}
		DrumSchema schema = loadDrumSchemaIfAvailable(drumSchemaFileName, drumRhythmName);
		if (schema != null) {
			cachedDrumSchema = schema;
			return true;
		}
		return false;
	}

	private DrumSchema loadDrumSchemaIfAvailable(String drumSchemaFileName, String drumRhythmName) {
		if (drumSchemaFileName == null || drumRhythmName == null) {
			return null;
		}
		File schemaFile = audioProject.getRelativeFile(drumSchemaFileName);
		if (!schemaFile.exists()) {
			return null;
		}
		try {
			DrumSchema schema = loadDrumSchema(schemaFile);
			DrumRhythm drumRhythm = schema.getDrumRhythmByName(drumRhythmName);
			if (drumRhythm != null) {
				return schema;
			}
			return null;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	void addDrumAudioDeviceListener(DrumAudioDeviceListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<DrumAudioDeviceListener>();
		}
		listeners.add(listener);
	}

	void removeDrumAudioDeviceListener(DrumAudioDeviceListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}

	String getDrumSchemaFileName() {
		return drumSchemaFileName;
	}

	String getDrumRhythmName() {
		return drumRhythmName;
	}
}

==> audiochain-1.1.6-src.tgz: org/audiochain/devices/drum/DrumAudioDevice.java <==
/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.drum;

import java.io.BufferedInputStream;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;

import javax.sound.sampled.UnsupportedAudioFileException;
import javax.xml.parsers.ParserConfigurationException;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.FrameObserver;
import org.audiochain.io.LimitingAudioDataReader;
import org.audiochain.model.AbstractAudioDevice;
import org.audiochain.model.AudioDeviceProperty;
import org.audiochain.model.BinaryAudioProject;
import org.audiochain.model.BinaryDataset;
import org.audiochain.model.BinaryFileDataset;
import org.audiochain.model.BinarySynchronizable;
import org.audiochain.model.Performable;
import org.audiochain.model.SourceAudioDevice;
import org.audiochain.model.Synchronizable;
import org.audiochain.model.UserInterfaceContext;
import org.xml.sax.SAXException;

/**
 * TODO remember all synced but unsaved files and remove them when the project
 * becomes closed without saving (analogous to recording device)
 */
public class DrumAudioDevice extends AbstractAudioDevice implements SourceAudioDevice, BinarySynchronizable {
	private static final long serialVersionUID = 1L;
	private static final String DRUM_SCHEMA_PROPERTY_NAME = "DrumSchema";
	private static final String DRUM_RHYTHM_PROPERTY_NAME = "Rhythm";
	private String drumSchemaFileName;
	private String drumRhythmName;
	@AudioDeviceProperty(name = "BPM", min = 20, max = 350, symbol = 9684, unit = "Bpm")
	private float beatsPerMinute = 80F;
	@AudioDeviceProperty(name = "Repeat", symbol = 10226, falseSymbol = 10230)
	private boolean repeat;
	private transient DrumAudioDataReader drumReader;
	private transient Collection<DrumAudioDeviceListener> listeners;
	private transient DrumSchema cachedDrumSchema;

	@Override
	public AudioDataReader createAudioDataReader(float frameRate, int sampleSizeInBits, FrameObserver frameObserver) {
		DrumSchema drumSchema = getDrumSchema();
		DrumRhythm rhythm = drumSchema.getDrumRhythmByName(drumRhythmName);
		drumReader = new DrumAudioDataReader(drumSchema, rhythm, beatsPerMinute, frameRate);
		drumReader.setRepeat(repeat);
		AudioDataReader reader = new LimitingAudioDataReader(drumReader, sampleSizeInBits);
		return reader;
	}

	/*
	 * This method is called from ui.
	 */
	void setDrumSchemaFile(File drumSchemaFile) throws FileNotFoundException, IOException, DrumSchemaFormatException, UnsupportedAudioFileException, SAXException, ParserConfigurationException {
		if (drumSchemaFile == null) {
			throw new IllegalArgumentException("Drum schema file must not be null.");
		}
		// throw exceptions if file is not valid
		DrumSchema schema = loadDrumSchema(drumSchemaFile);
		// update the new file change time and the ui
		String oldDrumSchemaFileName = this.drumSchemaFileName;
		String newDrumSchemaFileName = audioProject.getRelativePath(drumSchemaFile.getAbsolutePath());
		if (!isEqual(oldDrumSchemaFileName, newDrumSchemaFileName)) {
			this.drumSchemaFileName = newDrumSchemaFileName;
			setAudioDeviceValueLastChangeTime(DRUM_SCHEMA_PROPERTY_NAME, System.currentTimeMillis());
			// preselect the first rhythm for the new file
			Collection<String> drumRhythmNames = schema.getDrumRhythmNames();
			// a schema must have at least one rhytm
			this.drumRhythmName = drumRhythmNames.iterator().next();
			setAudioDeviceValueLastChangeTime(DRUM_RHYTHM_PROPERTY_NAME, System.currentTimeMillis());
			cachedDrumSchema = null;
			fireDrumSchemaFileNameChanged();
			fireDrumRhythmChanged();
			firePerformanceRelevantAudioDevicePropertyChanged();
			audioProject.changeFrameLength();
			// trigger the binary drum file synchronization
			fireBinaryChanged(new BinaryAudioProject(audioProject));
		}
	}

	private void fireDrumSchemaFileNameChanged() {
		if (listeners == null) {
			return;
		}
		for (DrumAudioDeviceListener listener : listeners) {
			listener.drumSchemaFileNameChanged(drumSchemaFileName);
		}
	}

	/*
	 * This method must be called only if isAudioDataAvailable() returns true.
	 */
	private DrumSchema getDrumSchema() {
		if (!isAudioDataAvailable()) {
			throw new IllegalStateException("Drum schema is not available.");
		}
		return cachedDrumSchema;
	}

	private DrumSchema loadDrumSchema(File drumSchemaFile) throws FileNotFoundException, IOException, DrumSchemaFormatException, UnsupportedAudioFileException, SAXException,
			ParserConfigurationException {
		DrumSchema ret;
		String projectPath = audioProject.getProjectPath();
		String schemaFileName = drumSchemaFile.getName();
		if (schemaFileName.endsWith(".h2song")) {
			FileInputStream fis = new FileInputStream(drumSchemaFile);
			BufferedInputStream bis = new BufferedInputStream(fis);
			Collection<File> clipSearchPaths = DrumSettings.getInstance().getClipSearchPaths();
			clipSearchPaths.add(new File(projectPath));
			try {
				HydrogenDrumSchemaReader hydrogenDrumSchemaReader = new HydrogenDrumSchemaReader(bis, clipSearchPaths);
				ret = hydrogenDrumSchemaReader.getSchema();
				File dump = new File(projectPath, schemaFileName.substring(0, schemaFileName.length() - 7) + "-gen.txt");
				BufferedWriter bw = new BufferedWriter(new FileWriter(dump));
				bw.write(ret.toString());
				bw.close();
			} finally {
				bis.close();
				fis.close();
			}
		} else {
			FileReader fr = new FileReader(drumSchemaFile);
			// search for clip files within the project and relative to the drum
			// schema file
			ret = DrumSchema.readDrumSchema(fr, new File(projectPath), drumSchemaFile.getParentFile());
			fr.close();
		}
		ret.setLastModified(drumSchemaFile.lastModified());
		return ret;
	}

	/* is called from ui */
	void setDrumRhythm(String drumRhythmName) {
		String oldDrumRhythmName = this.drumRhythmName;
		if (!isEqual(oldDrumRhythmName, drumRhythmName)) {
			this.drumRhythmName = drumRhythmName;
			setAudioDeviceValueLastChangeTime(DRUM_RHYTHM_PROPERTY_NAME, System.currentTimeMillis());
			updateReader();
			firePerformanceRelevantAudioDevicePropertyChanged();
			audioProject.changeFrameLength();
		}
	}

	private void updateReader() {
		if (drumReader != null && isAudioDataAvailable()) {
			DrumSchema drumSchema = getDrumSchema();
			DrumRhythm rhythm = drumSchema.getDrumRhythmByName(drumRhythmName);
			drumReader.setDrumSchemaAndRhythm(drumSchema, rhythm);
		}
	}

	private void fireDrumRhythmChanged() {
		if (listeners == null) {
			return;
		}
		for (DrumAudioDeviceListener listener : listeners) {
			listener.drumRhythmNameChanged(drumRhythmName);
		}
	}

	void setBeatsPerMinute(float beatPerMinute) {
		this.beatsPerMinute = beatPerMinute;
		if (drumReader != null) {
			drumReader.setPulsesPerMinute(beatPerMinute);
		}
		firePerformanceRelevantAudioDevicePropertyChanged();
		audioProject.changeFrameLength();
	}

	void setRepeat(boolean repeat) {
		this.repeat = repeat;
		if (drumReader != null) {
			drumReader.setRepeat(repeat);
		}
		firePerformanceRelevantAudioDevicePropertyChanged();
		audioProject.changeFrameLength();
	}

	Collection<String> getDrumRhythmNames() {
		Collection<String> ret = new LinkedHashSet<String>();
		if (isAudioDataAvailable()) {
			DrumSchema drumSchema = getDrumSchema();
			ret.addAll(drumSchema.getDrumRhythmNames());
		}
		return ret;
	}

	@Override
	public long getMaxFrame() {
		if (!isAudioDataAvailable()) {
			return -1;
		}
		DrumSchema drumSchema = getDrumSchema();
		DrumRhythm rhythm = drumSchema.getDrumRhythmByName(drumRhythmName);
		DrumAudioDataReader drumReader = new DrumAudioDataReader(drumSchema, rhythm, beatsPerMinute, audioProject.getFrameRate());
		drumReader.setRepeat(repeat);
		long rhythmFrameCount = drumReader.getRhythmFrameCount();
		drumReader.close();
		return rhythmFrameCount;
	}

	@Override
	protected UserInterfaceContext<DrumAudioDevice> createUserInterfaceContext() {
		return new DrumUserInterfaceContext();
	}

	@Override
	public boolean isRecordable() {
		return false;
	}

	@Override
	public Performable getPerformable() {
		return null;
	}

	@Override
	public String getName() {
		return "Drum";
	}

	@Override
	public Collection<BinaryDataset> getAllBinaryDatasets() {
		Collection<BinaryDataset> ret = new HashSet<BinaryDataset>();
		if (drumSchemaFileName != null) {
			File schemaFile = audioProject.getRelativeFile(drumSchemaFileName);
			if (schemaFile.exists()) {
				ret.add(new BinaryFileDataset(audioProject, schemaFile));
			}
		}
		if (isAudioDataAvailable()) {
			DrumSchema drumSchema = getDrumSchema();
			for (DrumClip clip : drumSchema.getClips()) {
				File clipFile = clip.getFile();
				if (clipFile.exists() && clipFile.canRead()) {
					ret.add(new BinaryFileDataset(audioProject, clipFile));
				}
			}
		}
		return ret;
	}

	@Override
	public void synchronizeBy(Synchronizable source) {
		// sync bpm and repeat
		super.synchronizeBy(source);

		if (source instanceof DrumAudioDevice) {
			DrumAudioDevice sourceDevice = (DrumAudioDevice) source;
			long sourceRhythmChangeTime = sourceDevice.getAudioDeviceValueLastChangeTime(DRUM_RHYTHM_PROPERTY_NAME);
			long rhythmChangeTime = getAudioDeviceValueLastChangeTime(DRUM_RHYTHM_PROPERTY_NAME);
			if (!isEqual(sourceDevice.drumSchemaFileName, drumSchemaFileName)) {
				long sourceSchemaChangeTime = sourceDevice.getAudioDeviceValueLastChangeTime(DRUM_SCHEMA_PROPERTY_NAME);
				long schemaChangeTime = getAudioDeviceValueLastChangeTime(DRUM_SCHEMA_PROPERTY_NAME);
				if (greaterOrBothZero(sourceSchemaChangeTime, schemaChangeTime)) {
					DrumSchema schema = loadDrumSchemaIfAvailable(sourceDevice.drumSchemaFileName, sourceDevice.drumRhythmName);
					if (schema != null) {
						drumSchemaFileName = sourceDevice.drumSchemaFileName;
						setAudioDeviceValueLastChangeTime(DRUM_SCHEMA_PROPERTY_NAME, sourceSchemaChangeTime);
						drumRhythmName = sourceDevice.drumRhythmName;
						setAudioDeviceValueLastChangeTime(DRUM_RHYTHM_PROPERTY_NAME, sourceRhythmChangeTime);
						cachedDrumSchema = schema;
						updateReader();
						fireDrumSchemaFileNameChanged();
						fireDrumRhythmChanged();
						firePerformanceRelevantAudioDevicePropertyChanged();
						audioProject.changeFrameLength();
					} else {
						// the schema changed, but is not known at this side
						drumSchemaFileName = null;
						cachedDrumSchema = null;
						// tell the other side that this side needs the binary files
						fireBinaryChanged(new BinaryAudioProject(audioProject));
					}
				}
			} else if (!isEqual(sourceDevice.drumRhythmName, drumRhythmName)) {
				if (greaterOrBothZero(sourceRhythmChangeTime, rhythmChangeTime)) {
					this.drumRhythmName = sourceDevice.drumRhythmName;
					setAudioDeviceValueLastChangeTime(DRUM_RHYTHM_PROPERTY_NAME, sourceRhythmChangeTime);
					updateReader();
					fireDrumRhythmChanged();
					firePerformanceRelevantAudioDevicePropertyChanged();
					audioProject.changeFrameLength();
				}
			}
		}
	}

	@Override
	public boolean isAudioDataAvailable() {
		if (cachedDrumSchema != null) {
			// reload schema file if changed
			if (getDrumSchemaFileLastModified() > cachedDrumSchema.getLastModified()) {
				cachedDrumSchema = loadDrumSchemaIfAvailable(drumSchemaFileName, drumRhythmName);
				assert cachedDrumSchema != null;
			}
			return true;
		}
		DrumSchema schema = loadDrumSchemaIfAvailable(drumSchemaFileName, drumRhythmName);
		if (schema != null) {
			cachedDrumSchema = schema;
			return true;
		}
		return false;
	}
	
	private long getDrumSchemaFileLastModified() {
		if (drumSchemaFileName != null) {
			File schemaFile = audioProject.getRelativeFile(drumSchemaFileName);
			if (schemaFile.exists()) {
				return schemaFile.lastModified();
			}
		}
		return -1;
	}

	private DrumSchema loadDrumSchemaIfAvailable(String drumSchemaFileName, String drumRhythmName) {
		if (drumSchemaFileName == null || drumRhythmName == null) {
			return null;
		}
		File schemaFile = audioProject.getRelativeFile(drumSchemaFileName);
		if (!schemaFile.exists()) {
			return null;
		}
		try {
			DrumSchema schema = loadDrumSchema(schemaFile);
			DrumRhythm drumRhythm = schema.getDrumRhythmByName(drumRhythmName);
			if (drumRhythm != null) {
				return schema;
			}
			return null;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	void addDrumAudioDeviceListener(DrumAudioDeviceListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<DrumAudioDeviceListener>();
		}
		listeners.add(listener);
	}

	void removeDrumAudioDeviceListener(DrumAudioDeviceListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}

	String getDrumSchemaFileName() {
		return drumSchemaFileName;
	}

	String getDrumRhythmName() {
		return drumRhythmName;
	}
}
